# The MIT License (MIT)
#
# Copyright (c) 2015-2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

_CLOCK_CB = '{return_ctype} (*{cname}_clock_get_value)(void *);'


_PLATFORM_CALLBACKS_BEGIN = '''/* barectf platform callbacks */
struct {prefix}platform_callbacks {{
	/* clock callbacks */'''


_PLATFORM_CALLBACKS_END = '''
	/* is back-end full? */
	int (*is_backend_full)(void *);

	/* open packet */
	void (*open_packet)(void *);

	/* close packet */
	void (*close_packet)(void *);
};'''


_CTX_PARENT = '''/* common barectf context */
struct {prefix}ctx {{
	/* platform callbacks */
	struct {prefix}platform_callbacks cbs;

	/* platform data (passed to callbacks) */
	void *data;

	/* output buffer (will contain a CTF binary packet) */
	uint8_t *buf;

	/* packet size in bits */
	uint32_t packet_size;

	/* content size in bits */
	uint32_t content_size;

	/* current position from beginning of packet in bits */
	uint32_t at;

	/* packet header + context size (content offset) */
	uint32_t off_content;

	/* events discarded */
	uint32_t events_discarded;

	/* current packet is opened */
	int packet_is_open;

	/* in tracing code */
	volatile int in_tracing_section;

	/* tracing is enabled */
	volatile int is_tracing_enabled;

	/* use current/last event timestamp when opening/closing packets */
	int use_cur_last_event_ts;
}};'''


_CTX_BEGIN = '''/* context for stream "{sname}" */
struct {prefix}{sname}_ctx {{
	/* parent */
	struct {prefix}ctx parent;

	/* config-specific members follow */'''


_CTX_END = '};'


_FUNC_INIT_PROTO = '''/* initialize context */
void {prefix}init(
	void *vctx,
	uint8_t *buf,
	uint32_t buf_size,
	struct {prefix}platform_callbacks cbs,
	void *data
)'''


_FUNC_INIT_BODY = '''{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);
	ctx->cbs = cbs;
	ctx->data = data;
	ctx->buf = buf;
	ctx->packet_size = _BYTES_TO_BITS(buf_size);
	ctx->at = 0;
	ctx->events_discarded = 0;
	ctx->packet_is_open = 0;
	ctx->in_tracing_section = 0;
	ctx->is_tracing_enabled = 1;
	ctx->use_cur_last_event_ts = 0;
}}'''


_FUNC_OPEN_PROTO_BEGIN = '''/* open packet for stream "{sname}" */
void {prefix}{sname}_open_packet(
	struct {prefix}{sname}_ctx *ctx'''


_FUNC_OPEN_PROTO_END = ')'


_FUNC_OPEN_BODY_BEGIN = '''{{
{ts}
	const int saved_in_tracing_section = ctx->parent.in_tracing_section;

	/*
	 * This function is either called by a tracing function, or
	 * directly by the platform.
	 *
	 * If it's called by a tracing function, then
	 * ctx->parent.in_tracing_section is 1, so it's safe to open
	 * the packet here (alter the packet), even if tracing was
	 * disabled in the meantime because we're already in a tracing
	 * section (which finishes at the end of the tracing function
	 * call).
	 *
	 * If it's called directly by the platform, then if tracing is
	 * disabled, we don't want to alter the packet, and return
	 * immediately.
	 */
	if (!ctx->parent.is_tracing_enabled && !saved_in_tracing_section) {{
		ctx->parent.in_tracing_section = 0;
		return;
	}}

	/* we can modify the packet */
	ctx->parent.in_tracing_section = 1;
'''


_FUNC_OPEN_BODY_END = '''
	/* save content beginning's offset */
	ctx->parent.off_content = ctx->parent.at;

	/* mark current packet as open */
	ctx->parent.packet_is_open = 1;

	/* not tracing anymore */
	ctx->parent.in_tracing_section = saved_in_tracing_section;
}'''


_FUNC_CLOSE_PROTO = '''/* close packet for stream "{sname}" */
void {prefix}{sname}_close_packet(struct {prefix}{sname}_ctx *ctx)'''


_FUNC_CLOSE_BODY_BEGIN = '''{{
{ts}
	const int saved_in_tracing_section = ctx->parent.in_tracing_section;

	/*
	 * This function is either called by a tracing function, or
	 * directly by the platform.
	 *
	 * If it's called by a tracing function, then
	 * ctx->parent.in_tracing_section is 1, so it's safe to close
	 * the packet here (alter the packet), even if tracing was
	 * disabled in the meantime, because we're already in a tracing
	 * section (which finishes at the end of the tracing function
	 * call).
	 *
	 * If it's called directly by the platform, then if tracing is
	 * disabled, we don't want to alter the packet, and return
	 * immediately.
	 */
	if (!ctx->parent.is_tracing_enabled && !saved_in_tracing_section) {{
		ctx->parent.in_tracing_section = 0;
		return;
	}}

	/* we can modify the packet */
	ctx->parent.in_tracing_section = 1;
'''


_FUNC_CLOSE_BODY_END = '''
	/* go back to end of packet */
	ctx->parent.at = ctx->parent.packet_size;

	/* mark packet as closed */
	ctx->parent.packet_is_open = 0;

	/* not tracing anymore */
	ctx->parent.in_tracing_section = saved_in_tracing_section;
}'''


_DEFINE_DEFAULT_STREAM_TRACE = '#define {prefix}trace_{evname} {prefix}{sname}_trace_{evname}'


_FUNC_TRACE_PROTO_BEGIN = '''/* trace (stream "{sname}", event "{evname}") */
void {prefix}{sname}_trace_{evname}(
	struct {prefix}{sname}_ctx *ctx'''


_FUNC_TRACE_PROTO_END = ')'


_FUNC_TRACE_BODY = '''{{
	uint32_t ev_size;

	/* save timestamp */
	{save_ts}

	if (!ctx->parent.is_tracing_enabled) {{
		return;
	}}

	/* we can modify the packet */
	ctx->parent.in_tracing_section = 1;

	/* get event size */
	ev_size = _get_event_size_{sname}_{evname}(TO_VOID_PTR(ctx){params});

	/* do we have enough space to serialize? */
	if (!_reserve_event_space(TO_VOID_PTR(ctx), ev_size)) {{
		/* no: forget this */
		ctx->parent.in_tracing_section = 0;
		return;
	}}

	/* serialize event */
	_serialize_event_{sname}_{evname}(TO_VOID_PTR(ctx){params});

	/* commit event */
	_commit_event(TO_VOID_PTR(ctx));

	/* not tracing anymore */
	ctx->parent.in_tracing_section = 0;
}}'''


_FUNC_GET_EVENT_SIZE_PROTO_BEGIN = '''static uint32_t _get_event_size_{sname}_{evname}(
	void *vctx'''


_FUNC_GET_EVENT_SIZE_PROTO_END = ')'


_FUNC_GET_EVENT_SIZE_BODY_BEGIN = '''{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);
	uint32_t at = ctx->at;'''


_FUNC_GET_EVENT_SIZE_BODY_END = '''	return at - ctx->at;
}'''


_FUNC_SERIALIZE_STREAM_EVENT_HEADER_PROTO_BEGIN = '''static void _serialize_stream_event_header_{sname}(
	void *vctx,
	uint32_t event_id'''


_FUNC_SERIALIZE_STREAM_EVENT_HEADER_PROTO_END = ')'


_FUNC_SERIALIZE_STREAM_EVENT_HEADER_BODY_BEGIN = '''{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);'''


_FUNC_SERIALIZE_STREAM_EVENT_HEADER_BODY_END = '}'


_FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_PROTO_BEGIN = '''static void _serialize_stream_event_context_{sname}(
	void *vctx'''


_FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_PROTO_END = ')'


_FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_BODY_BEGIN = '''{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);'''


_FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_BODY_END = '}'


_FUNC_SERIALIZE_EVENT_PROTO_BEGIN = '''static void _serialize_event_{sname}_{evname}(
	void *vctx'''


_FUNC_SERIALIZE_EVENT_PROTO_END = ')'


_FUNC_SERIALIZE_EVENT_BODY_BEGIN = '''{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);'''


_FUNC_SERIALIZE_EVENT_BODY_END = '}'


_HEADER_BEGIN = '''#ifndef _{ucprefix}H
#define _{ucprefix}H

/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2015-2016 Philippe Proulx <pproulx@efficios.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *
 * The following C code was generated by barectf {version}
 * on {date}.
 *
 * For more details, see <http://barectf.org>.
 */

#include <stdint.h>

#include "{bitfield_header_filename}"

#ifdef __cplusplus
extern "C" {{
#endif

{prefix_def}
{default_stream_def}

{default_stream_trace_defs}

struct {prefix}ctx;

uint32_t {prefix}packet_size(void *ctx);
int {prefix}packet_is_full(void *ctx);
int {prefix}packet_is_empty(void *ctx);
uint32_t {prefix}packet_events_discarded(void *ctx);
uint8_t *{prefix}packet_buf(void *ctx);
void {prefix}packet_set_buf(void *ctx, uint8_t *buf, uint32_t buf_size);
uint32_t {prefix}packet_buf_size(void *ctx);
int {prefix}packet_is_open(void *ctx);
int {prefix}is_in_tracing_section(void *ctx);
volatile const int *{prefix}is_in_tracing_section_ptr(void *ctx);
int {prefix}is_tracing_enabled(void *ctx);
void {prefix}enable_tracing(void *ctx, int enable);'''


_HEADER_END = '''#ifdef __cplusplus
}}
#endif

#endif /* _{ucprefix}H */
'''


_C_SRC = '''/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2015-2016 Philippe Proulx <pproulx@efficios.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *
 * The following C code was generated by barectf {version}
 * on {date}.
 *
 * For more details, see <http://barectf.org>.
 */

#include <stdint.h>
#include <string.h>
#include <assert.h>

#include "{header_filename}"

#define _ALIGN(_at, _align)					\\
	do {{							\\
		(_at) = ((_at) + ((_align) - 1)) & -(_align);	\\
	}} while (0)

#ifdef __cplusplus
# define TO_VOID_PTR(_value)		static_cast<void *>(_value)
# define FROM_VOID_PTR(_type, _value)	static_cast<_type *>(_value)
#else
# define TO_VOID_PTR(_value)		((void *) (_value))
# define FROM_VOID_PTR(_type, _value)	((_type *) (_value))
#endif

#define _BITS_TO_BYTES(_x)	((_x) >> 3)
#define _BYTES_TO_BITS(_x)	((_x) << 3)

union f2u {{
	float f;
	uint32_t u;
}};

union d2u {{
	double f;
	uint64_t u;
}};

uint32_t {prefix}packet_size(void *ctx)
{{
	return FROM_VOID_PTR(struct {prefix}ctx, ctx)->packet_size;
}}

int {prefix}packet_is_full(void *ctx)
{{
	struct {prefix}ctx *cctx = FROM_VOID_PTR(struct {prefix}ctx, ctx);

	return cctx->at == cctx->packet_size;
}}

int {prefix}packet_is_empty(void *ctx)
{{
	struct {prefix}ctx *cctx = FROM_VOID_PTR(struct {prefix}ctx, ctx);

	return cctx->at <= cctx->off_content;
}}

uint32_t {prefix}packet_events_discarded(void *ctx)
{{
	return FROM_VOID_PTR(struct {prefix}ctx, ctx)->events_discarded;
}}

uint8_t *{prefix}packet_buf(void *ctx)
{{
	return FROM_VOID_PTR(struct {prefix}ctx, ctx)->buf;
}}

uint32_t {prefix}packet_buf_size(void *ctx)
{{
	struct {prefix}ctx *cctx = FROM_VOID_PTR(struct {prefix}ctx, ctx);

	return _BITS_TO_BYTES(cctx->packet_size);
}}

void {prefix}packet_set_buf(void *ctx, uint8_t *buf, uint32_t buf_size)
{{
	struct {prefix}ctx *cctx = FROM_VOID_PTR(struct {prefix}ctx, ctx);

	cctx->buf = buf;
	cctx->packet_size = _BYTES_TO_BITS(buf_size);
}}

int {prefix}packet_is_open(void *ctx)
{{
	return FROM_VOID_PTR(struct {prefix}ctx, ctx)->packet_is_open;
}}

int {prefix}is_in_tracing_section(void *ctx)
{{
	return FROM_VOID_PTR(struct {prefix}ctx, ctx)->in_tracing_section;
}}

volatile const int *{prefix}is_in_tracing_section_ptr(void *ctx)
{{
	return &FROM_VOID_PTR(struct {prefix}ctx, ctx)->in_tracing_section;
}}

int {prefix}is_tracing_enabled(void *ctx)
{{
	return FROM_VOID_PTR(struct {prefix}ctx, ctx)->is_tracing_enabled;
}}

void {prefix}enable_tracing(void *ctx, int enable)
{{
	FROM_VOID_PTR(struct {prefix}ctx, ctx)->is_tracing_enabled = enable;
}}

static
void _write_cstring(struct {prefix}ctx *ctx, const char *src)
{{
	uint32_t sz = strlen(src) + 1;

	memcpy(&ctx->buf[_BITS_TO_BYTES(ctx->at)], src, sz);
	ctx->at += _BYTES_TO_BITS(sz);
}}

static
int _reserve_event_space(void *vctx, uint32_t ev_size)
{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);

	/* event _cannot_ fit? */
	if (ev_size > (ctx->packet_size - ctx->off_content)) {{
		ctx->events_discarded++;

		return 0;
	}}

	/* packet is full? */
	if ({prefix}packet_is_full(ctx)) {{
		/* yes: is back-end full? */
		if (ctx->cbs.is_backend_full(ctx->data)) {{
			/* yes: discard event */
			ctx->events_discarded++;

			return 0;
		}}

		/* back-end is not full: open new packet */
		ctx->use_cur_last_event_ts = 1;
		ctx->cbs.open_packet(ctx->data);
		ctx->use_cur_last_event_ts = 0;
	}}

	/* event fits the current packet? */
	if (ev_size > (ctx->packet_size - ctx->at)) {{
		/* no: close packet now */
		ctx->use_cur_last_event_ts = 1;
		ctx->cbs.close_packet(ctx->data);
		ctx->use_cur_last_event_ts = 0;

		/* is back-end full? */
		if (ctx->cbs.is_backend_full(ctx->data)) {{
			/* yes: discard event */
			ctx->events_discarded++;

			return 0;
		}}

		/* back-end is not full: open new packet */
		ctx->use_cur_last_event_ts = 1;
		ctx->cbs.open_packet(ctx->data);
		ctx->use_cur_last_event_ts = 0;
		assert(ev_size <= (ctx->packet_size - ctx->at));
	}}

	return 1;
}}

static
void _commit_event(void *vctx)
{{
	struct {prefix}ctx *ctx = FROM_VOID_PTR(struct {prefix}ctx, vctx);

	/* is packet full? */
	if ({prefix}packet_is_full(ctx)) {{
		/* yes: close it now */
		ctx->cbs.close_packet(ctx->data);
	}}
}}'''


_BITFIELD = '''#ifndef _$PREFIX$BITFIELD_H
#define _$PREFIX$BITFIELD_H

/*
 * BabelTrace
 *
 * Bitfields read/write functions.
 *
 * Copyright 2010 - Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <limits.h>

#ifdef __cplusplus
# define CAST_PTR(_type, _value) \\
	static_cast<_type>(static_cast<void *>(_value))
#else
# define CAST_PTR(_type, _value)	((void *) (_value))
#endif

#define $PREFIX$BYTE_ORDER $ENDIAN_DEF$

/* We can't shift a int from 32 bit, >> 32 and << 32 on int is undefined */
#define _$prefix$bt_piecewise_rshift(_vtype, _v, _shift) \\
do {									\\
	unsigned long ___shift = (_shift);				\\
	unsigned long sb = (___shift) / (sizeof(_v) * CHAR_BIT - 1);	\\
	unsigned long final = (___shift) % (sizeof(_v) * CHAR_BIT - 1); \\
									\\
	for (; sb; sb--)						\\
		_v >>= sizeof(_v) * CHAR_BIT - 1;			\\
	_v >>= final;							\\
} while (0)

/*
 * $prefix$bt_bitfield_write - write integer to a bitfield in native endianness
 *
 * Save integer to the bitfield, which starts at the "start" bit, has "len"
 * bits.
 * The inside of a bitfield is from high bits to low bits.
 * Uses native endianness.
 * For unsigned "v", pad MSB with 0 if bitfield is larger than v.
 * For signed "v", sign-extend v if bitfield is larger than v.
 *
 * On little endian, bytes are placed from the less significant to the most
 * significant. Also, consecutive bitfields are placed from lower bits to higher
 * bits.
 *
 * On big endian, bytes are places from most significant to less significant.
 * Also, consecutive bitfields are placed from higher to lower bits.
 */

#define _$prefix$bt_bitfield_write_le(_ptr, type, _start, _length, _vtype, _v) \\
do {									\\
	_vtype __v = (_v);					\\
	type *__ptr = CAST_PTR(type *, _ptr);				\\
	unsigned long __start = (_start), __length = (_length);		\\
	type mask, cmask;						\\
	unsigned long ts = sizeof(type) * CHAR_BIT; /* type size */	\\
	unsigned long start_unit, end_unit, this_unit;			\\
	unsigned long end, cshift; /* cshift is "complement shift" */	\\
									\\
	if (!__length)							\\
		break;							\\
									\\
	end = __start + __length;					\\
	start_unit = __start / ts;					\\
	end_unit = (end + (ts - 1)) / ts;				\\
									\\
	/* Trim v high bits */						\\
	if (__length < sizeof(__v) * CHAR_BIT)				\\
		__v &= ~((~(_vtype) 0) << __length);		\\
									\\
	/* We can now append v with a simple "or", shift it piece-wise */ \\
	this_unit = start_unit;						\\
	if (start_unit == end_unit - 1) {				\\
		mask = ~((~(type) 0) << (__start % ts));		\\
		if (end % ts)						\\
			mask |= (~(type) 0) << (end % ts);		\\
		cmask = (type) __v << (__start % ts);			\\
		cmask &= ~mask;						\\
		__ptr[this_unit] &= mask;				\\
		__ptr[this_unit] |= cmask;				\\
		break;							\\
	}								\\
	if (__start % ts) {						\\
		cshift = __start % ts;					\\
		mask = ~((~(type) 0) << cshift);			\\
		cmask = (type) __v << cshift;				\\
		cmask &= ~mask;						\\
		__ptr[this_unit] &= mask;				\\
		__ptr[this_unit] |= cmask;				\\
		_$prefix$bt_piecewise_rshift(_vtype, __v, ts - cshift); \\
		__start += ts - cshift;					\\
		this_unit++;						\\
	}								\\
	for (; this_unit < end_unit - 1; this_unit++) {			\\
		__ptr[this_unit] = (type) __v;				\\
		_$prefix$bt_piecewise_rshift(_vtype, __v, ts); 		\\
		__start += ts;						\\
	}								\\
	if (end % ts) {							\\
		mask = (~(type) 0) << (end % ts);			\\
		cmask = (type) __v;					\\
		cmask &= ~mask;						\\
		__ptr[this_unit] &= mask;				\\
		__ptr[this_unit] |= cmask;				\\
	} else								\\
		__ptr[this_unit] = (type) __v;				\\
} while (0)

#define _$prefix$bt_bitfield_write_be(_ptr, type, _start, _length, _vtype, _v) \\
do {									\\
	_vtype __v = (_v);					\\
	type *__ptr = CAST_PTR(type *, _ptr);				\\
	unsigned long __start = (_start), __length = (_length);		\\
	type mask, cmask;						\\
	unsigned long ts = sizeof(type) * CHAR_BIT; /* type size */	\\
	unsigned long start_unit, end_unit, this_unit;			\\
	unsigned long end, cshift; /* cshift is "complement shift" */	\\
									\\
	if (!__length)							\\
		break;							\\
									\\
	end = __start + __length;					\\
	start_unit = __start / ts;					\\
	end_unit = (end + (ts - 1)) / ts;				\\
									\\
	/* Trim v high bits */						\\
	if (__length < sizeof(__v) * CHAR_BIT)				\\
		__v &= ~((~(_vtype) 0) << __length);			\\
									\\
	/* We can now append v with a simple "or", shift it piece-wise */ \\
	this_unit = end_unit - 1;					\\
	if (start_unit == end_unit - 1) {				\\
		mask = ~((~(type) 0) << ((ts - (end % ts)) % ts));	\\
		if (__start % ts)					\\
			mask |= (~((type) 0)) << (ts - (__start % ts));	\\
		cmask = (type) __v << ((ts - (end % ts)) % ts);		\\
		cmask &= ~mask;						\\
		__ptr[this_unit] &= mask;				\\
		__ptr[this_unit] |= cmask;				\\
		break;							\\
	}								\\
	if (end % ts) {							\\
		cshift = end % ts;					\\
		mask = ~((~(type) 0) << (ts - cshift));			\\
		cmask = (type) __v << (ts - cshift);			\\
		cmask &= ~mask;						\\
		__ptr[this_unit] &= mask;				\\
		__ptr[this_unit] |= cmask;				\\
		_$prefix$bt_piecewise_rshift(_vtype, __v, cshift); \\
		end -= cshift;						\\
		this_unit--;						\\
	}								\\
	for (; (long) this_unit >= (long) start_unit + 1; this_unit--) { \\
		__ptr[this_unit] = (type) __v;				\\
		_$prefix$bt_piecewise_rshift(_vtype, __v, ts); \\
		end -= ts;						\\
	}								\\
	if (__start % ts) {						\\
		mask = (~(type) 0) << (ts - (__start % ts));		\\
		cmask = (type) __v;					\\
		cmask &= ~mask;						\\
		__ptr[this_unit] &= mask;				\\
		__ptr[this_unit] |= cmask;				\\
	} else								\\
		__ptr[this_unit] = (type) __v;				\\
} while (0)

/*
 * $prefix$bt_bitfield_write_le - write integer to a bitfield in little endian
 * $prefix$bt_bitfield_write_be - write integer to a bitfield in big endian
 */

#if ($PREFIX$BYTE_ORDER == LITTLE_ENDIAN)

#define $prefix$bt_bitfield_write_le(ptr, type, _start, _length, _vtype, _v) \\
	_$prefix$bt_bitfield_write_le(ptr, type, _start, _length, _vtype, _v)

#define $prefix$bt_bitfield_write_be(ptr, type, _start, _length, _vtype, _v) \\
	_$prefix$bt_bitfield_write_be(ptr, unsigned char, _start, _length, _vtype, _v)

#elif ($PREFIX$BYTE_ORDER == BIG_ENDIAN)

#define $prefix$bt_bitfield_write_le(ptr, type, _start, _length, _vtype, _v) \\
	_$prefix$bt_bitfield_write_le(ptr, unsigned char, _start, _length, _vtype, _v)

#define $prefix$bt_bitfield_write_be(ptr, type, _start, _length, _vtype, _v) \\
	_$prefix$bt_bitfield_write_be(ptr, type, _start, _length, _vtype, _v)

#else /* ($PREFIX$BYTE_ORDER == PDP_ENDIAN) */

#error "Byte order not supported"

#endif

#endif /* _$PREFIX$BITFIELD_H */
'''



# The MIT License (MIT)
#
# Copyright (c) 2015 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from barectf import metadata
from barectf import codegen
import datetime
import barectf


_bo_to_string_map = {
    metadata.ByteOrder.LE: 'le',
    metadata.ByteOrder.BE: 'be',
}


_encoding_to_string_map = {
    metadata.Encoding.NONE: 'none',
    metadata.Encoding.ASCII: 'ASCII',
    metadata.Encoding.UTF8: 'UTF8',
}


def _bo_to_string(bo):
    return _bo_to_string_map[bo]


def _encoding_to_string(encoding):
    return _encoding_to_string_map[encoding]


def _bool_to_string(b):
    return 'true' if b else 'false'


def _gen_integer(t, cg):
    cg.add_line('integer {')
    cg.indent()
    cg.add_line('size = {};'.format(t.size))
    cg.add_line('align = {};'.format(t.align))
    cg.add_line('signed = {};'.format(_bool_to_string(t.signed)))
    cg.add_line('byte_order = {};'.format(_bo_to_string(t.byte_order)))
    cg.add_line('base = {};'.format(t.base))
    cg.add_line('encoding = {};'.format(_encoding_to_string(t.encoding)))

    if t.property_mappings:
        clock_name = t.property_mappings[0].object.name
        cg.add_line('map = clock.{}.value;'.format(clock_name))

    cg.unindent()
    cg.add_line('}')


def _gen_float(t, cg):
    cg.add_line('floating_point {')
    cg.indent()
    cg.add_line('exp_dig = {};'.format(t.exp_size))
    cg.add_line('mant_dig = {};'.format(t.mant_size))
    cg.add_line('align = {};'.format(t.align))
    cg.add_line('byte_order = {};'.format(_bo_to_string(t.byte_order)))
    cg.unindent()
    cg.add_line('}')


def _gen_enum(t, cg):
    cg.add_line('enum : ')
    cg.add_glue()
    _gen_type(t.value_type, cg)
    cg.append_to_last_line(' {')
    cg.indent()

    for label, (mn, mx) in t.members.items():
        if mn == mx:
            rg = str(mn)
        else:
            rg = '{} ... {}'.format(mn, mx)

        line = '"{}" = {},'.format(label, rg)
        cg.add_line(line)

    cg.unindent()
    cg.add_line('}')


def _gen_string(t, cg):
    cg.add_line('string {')
    cg.indent()
    cg.add_line('encoding = {};'.format(_encoding_to_string(t.encoding)))
    cg.unindent()
    cg.add_line('}')


def _find_deepest_array_element_type(t):
    if type(t) is metadata.Array:
        return _find_deepest_array_element_type(t.element_type)

    return t


def _fill_array_lengths(t, lengths):
    if type(t) is metadata.Array:
        lengths.append(t.length)
        _fill_array_lengths(t.element_type, lengths)


def _gen_struct_variant_entry(name, t, cg):
    real_t = _find_deepest_array_element_type(t)
    _gen_type(real_t, cg)
    cg.append_to_last_line(' {}'.format(name))

    # array
    lengths = []
    _fill_array_lengths(t, lengths)

    if lengths:
        for length in reversed(lengths):
            cg.append_to_last_line('[{}]'.format(length))

    cg.append_to_last_line(';')


def _gen_struct(t, cg):
    cg.add_line('struct {')
    cg.indent()

    for field_name, field_type in t.fields.items():
        _gen_struct_variant_entry(field_name, field_type, cg)

    cg.unindent()

    if not t.fields:
        cg.add_glue()

    cg.add_line('}} align({})'.format(t.min_align))


def _gen_variant(t, cg):
    cg.add_line('variant <{}> {{'.format(t.tag))
    cg.indent()

    for type_name, type_type in t.types.items():
        _gen_struct_variant_entry(type_name, type_type, cg)

    cg.unindent()

    if not t.types:
        cg.add_glue()

    cg.add_line('}')


_type_to_gen_type_func = {
    metadata.Integer: _gen_integer,
    metadata.FloatingPoint: _gen_float,
    metadata.Enum: _gen_enum,
    metadata.String: _gen_string,
    metadata.Struct: _gen_struct,
    metadata.Variant: _gen_variant,
}


def _gen_type(t, cg):
    _type_to_gen_type_func[type(t)](t, cg)


def _gen_entity(name, t, cg):
    cg.add_line('{} := '.format(name))
    cg.add_glue()
    _gen_type(t, cg)
    cg.append_to_last_line(';')


def _gen_start_block(name, cg):
    cg.add_line('{} {{'.format(name))
    cg.indent()


def _gen_end_block(cg):
    cg.unindent()
    cg.add_line('};')
    cg.add_empty_line()


def _gen_trace_block(meta, cg):
    trace = meta.trace

    _gen_start_block('trace', cg)
    cg.add_line('major = 1;')
    cg.add_line('minor = 8;')
    line = 'byte_order = {};'.format(_bo_to_string(trace.byte_order))
    cg.add_line(line)

    if trace.uuid is not None:
        line = 'uuid = "{}";'.format(trace.uuid)
        cg.add_line(line)

    if trace.packet_header_type is not None:
        _gen_entity('packet.header', trace.packet_header_type, cg)

    _gen_end_block(cg)


def _escape_literal_string(s):
    esc = s.replace('\\', '\\\\')
    esc = esc.replace('\n', '\\n')
    esc = esc.replace('\r', '\\r')
    esc = esc.replace('\t', '\\t')
    esc = esc.replace('"', '\\"')

    return esc


def _gen_env_block(meta, cg):
    env = meta.env

    if not env:
        return

    _gen_start_block('env', cg)

    for name, value in env.items():
        if type(value) is int:
            value_string = str(value)
        else:
            value_string = '"{}"'.format(_escape_literal_string(value))

        cg.add_line('{} = {};'.format(name, value_string))

    _gen_end_block(cg)


def _gen_clock_block(clock, cg):
    _gen_start_block('clock', cg)
    cg.add_line('name = {};'.format(clock.name))

    if clock.description is not None:
        desc = _escape_literal_string(clock.description)
        cg.add_line('description = "{}";'.format(desc))

    if clock.uuid is not None:
        cg.add_line('uuid = "{}";'.format(clock.uuid))

    cg.add_line('freq = {};'.format(clock.freq))
    cg.add_line('offset_s = {};'.format(clock.offset_seconds))
    cg.add_line('offset = {};'.format(clock.offset_cycles))
    cg.add_line('precision = {};'.format(clock.error_cycles))
    cg.add_line('absolute = {};'.format(_bool_to_string(clock.absolute)))
    _gen_end_block(cg)


def _gen_clock_blocks(meta, cg):
    clocks = meta.clocks

    for clock in clocks.values():
        _gen_clock_block(clock, cg)


def _gen_stream_block(meta, stream, cg):
    cg.add_cc_line(stream.name.replace('/', ''))
    _gen_start_block('stream', cg)

    if meta.trace.packet_header_type is not None:
        if 'stream_id' in meta.trace.packet_header_type.fields:
            cg.add_line('id = {};'.format(stream.id))

    if stream.packet_context_type is not None:
        _gen_entity('packet.context', stream.packet_context_type, cg)

    if stream.event_header_type is not None:
        _gen_entity('event.header', stream.event_header_type, cg)

    if stream.event_context_type is not None:
        _gen_entity('event.context', stream.event_context_type, cg)

    _gen_end_block(cg)


def _gen_event_block(meta, stream, ev, cg):
    _gen_start_block('event', cg)
    cg.add_line('name = "{}";'.format(ev.name))
    cg.add_line('id = {};'.format(ev.id))

    if meta.trace.packet_header_type is not None:
        if 'stream_id' in meta.trace.packet_header_type.fields:
            cg.add_line('stream_id = {};'.format(stream.id))

    cg.append_cc_to_last_line(stream.name.replace('/', ''))

    if ev.log_level is not None:
        add_fmt = ''

        if ev.log_level.name is not None:
            name = ev.log_level.name.replace('*/', '')
            add_fmt = ' /* {} */'.format(name)

        fmt = 'loglevel = {};' + add_fmt
        cg.add_line(fmt.format(ev.log_level.value))

    if ev.context_type is not None:
        _gen_entity('context', ev.context_type, cg)

    if ev.payload_type is not None:
        _gen_entity('fields', ev.payload_type, cg)
    else:
        fake_payload = metadata.Struct()
        fake_payload.min_align = 8
        _gen_entity('fields', fake_payload, cg)

    _gen_end_block(cg)


def _gen_streams_events_blocks(meta, cg):
    for stream in meta.streams.values():
        _gen_stream_block(meta, stream, cg)

        for ev in stream.events.values():
            _gen_event_block(meta, stream, ev, cg)


def from_metadata(meta):
    cg = codegen.CodeGenerator('\t')

    # version/magic
    cg.add_line('/* CTF 1.8 */')
    cg.add_empty_line()
    cg.add_line('''/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2015-2016 Philippe Proulx <pproulx@efficios.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 *''')
    v = barectf.__version__
    line = ' * The following TSDL code was generated by barectf v{}'.format(v)
    cg.add_line(line)
    now = datetime.datetime.now()
    line = ' * on {}.'.format(now)
    cg.add_line(line)
    cg.add_line(' *')
    cg.add_line(' * For more details, see <http://barectf.org>.')
    cg.add_line(' */')
    cg.add_empty_line()

    # trace block
    _gen_trace_block(meta, cg)

    # environment
    _gen_env_block(meta, cg)

    # clocks
    _gen_clock_blocks(meta, cg)

    # streams and contained events
    _gen_streams_events_blocks(meta, cg)

    return cg.code



# The MIT License (MIT)
#
# Copyright (c) 2015 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import enum
import collections


@enum.unique
class ByteOrder(enum.Enum):
    LE = 0
    BE = 1


@enum.unique
class Encoding(enum.Enum):
    NONE = 0
    UTF8 = 1
    ASCII = 2


class Type:
    @property
    def align(self):
        return None

    @property
    def size(self):
        pass

    @property
    def is_dynamic(self):
        raise NotImplementedError()


class PropertyMapping:
    def __init__(self, object, prop):
        self._object = object
        self._prop = prop

    @property
    def object(self):
        return self._object

    @object.setter
    def object(self, value):
        self._object = value

    @property
    def prop(self):
        return self.prop

    @prop.setter
    def prop(self, value):
        self.prop = value


class Integer(Type):
    def __init__(self):
        self.set_default_size()
        self.set_default_align()
        self.set_default_signed()
        self.set_default_byte_order()
        self.set_default_base()
        self.set_default_encoding()
        self.set_default_property_mappings()

    def set_default_size(self):
        self._size = None

    def set_default_align(self):
        self._align = None

    def set_default_signed(self):
        self._signed = False

    def set_default_byte_order(self):
        self._byte_order = None

    def set_default_base(self):
        self._base = 10

    def set_default_encoding(self):
        self._encoding = Encoding.NONE

    def set_default_property_mappings(self):
        self._property_mappings = []

    @property
    def signed(self):
        return self._signed

    @signed.setter
    def signed(self, value):
        self._signed = value

    @property
    def byte_order(self):
        return self._byte_order

    @byte_order.setter
    def byte_order(self, value):
        self._byte_order = value

    @property
    def base(self):
        return self._base

    @base.setter
    def base(self, value):
        self._base = value

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def align(self):
        if self._align is None:
            if self._size is None:
                return
            else:
                if self._size % 8 == 0:
                    return 8
                else:
                    return 1
        else:
            return self._align

    @align.setter
    def align(self, value):
        self._align = value

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    @property
    def property_mappings(self):
        return self._property_mappings

    @property
    def is_dynamic(self):
        return False


class FloatingPoint(Type):
    def __init__(self):
        self.set_default_exp_size()
        self.set_default_mant_size()
        self.set_default_align()
        self.set_default_byte_order()

    def set_default_exp_size(self):
        self._exp_size = None

    def set_default_mant_size(self):
        self._mant_size = None

    def set_default_align(self):
        self._align = 8

    def set_default_byte_order(self):
        self._byte_order = None

    @property
    def exp_size(self):
        return self._exp_size

    @exp_size.setter
    def exp_size(self, value):
        self._exp_size = value

    @property
    def mant_size(self):
        return self._mant_size

    @mant_size.setter
    def mant_size(self, value):
        self._mant_size = value

    @property
    def size(self):
        if self._exp_size is None or self._mant_size is None:
            return

        return self._exp_size + self._mant_size

    @property
    def byte_order(self):
        return self._byte_order

    @byte_order.setter
    def byte_order(self, value):
        self._byte_order = value

    @property
    def align(self):
        return self._align

    @align.setter
    def align(self, value):
        self._align = value

    @property
    def is_dynamic(self):
        return False


class Enum(Type):
    def __init__(self):
        self.set_default_value_type()
        self.set_default_members()

    def set_default_value_type(self):
        self._value_type = None

    def set_default_members(self):
        self._members = collections.OrderedDict()

    @property
    def align(self):
        return self._value_type.align

    @property
    def size(self):
        return self._value_type.size

    @property
    def value_type(self):
        return self._value_type

    @value_type.setter
    def value_type(self, value):
        self._value_type = value

    @property
    def members(self):
        return self._members

    @property
    def last_value(self):
        if not self._members:
            return

        return list(self._members.values())[-1][1]

    def value_of(self, label):
        return self._members[label]

    def label_of(self, value):
        for label, vrange in self._members.items():
            if value >= vrange[0] and value <= vrange[1]:
                return label

    def __getitem__(self, key):
        if type(key) is str:
            return self.value_of(key)
        elif type(key) is int:
            return self.label_of(key)

        raise TypeError('wrong subscript type')

    @property
    def is_dynamic(self):
        return False


class String(Type):
    def __init__(self):
        self.set_default_encoding()

    def set_default_encoding(self):
        self._encoding = Encoding.UTF8

    @property
    def align(self):
        return 8

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        self._encoding = value

    @property
    def is_dynamic(self):
        return True


class Array(Type):
    def __init__(self):
        self.set_default_element_type()
        self.set_default_length()

    def set_default_element_type(self):
        self._default_element_type = None

    def set_default_length(self):
        self._default_length = None

    @property
    def align(self):
        return self._element_type.align

    @property
    def element_type(self):
        return self._element_type

    @element_type.setter
    def element_type(self, value):
        self._element_type = value

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        self._length = value

    @property
    def is_variable_length(self):
        return type(self._length) is not int

    @property
    def is_dynamic(self):
        if self.is_variable_length:
            return True

        return self.element_type.is_dynamic


class Struct(Type):
    def __init__(self):
        self.set_default_min_align()
        self.set_default_fields()

    def set_default_min_align(self):
        self._min_align = 1

    def set_default_fields(self):
        self._fields = collections.OrderedDict()

    @property
    def min_align(self):
        return self._min_align

    @min_align.setter
    def min_align(self, value):
        self._min_align = value

    @property
    def align(self):
        align = self.min_align

        for field in self.fields.values():
            if field.align is None:
                return

            if field.align > align:
                align = field.align

        return align

    @property
    def is_dynamic(self):
        for field in self.fields.values():
            if field.is_dynamic:
                return True

        return False

    @property
    def fields(self):
        return self._fields

    def __getitem__(self, key):
        return self.fields[key]

    def __len__(self):
        return len(self._fields)


class Variant(Type):
    def __init__(self):
        self.set_default_tag()
        self.set_default_types()

    def set_default_tag(self):
        self._tag = None

    def set_default_types(self):
        self._types = collections.OrderedDict()

    @property
    def align(self):
        single_type = self.get_single_type()

        if single_type is not None:
            return single_type.align

        # if all the possible types have the same alignment, then
        # there's only one possible alignment
        align = None

        for t in self.types.values():
            if t.align is None:
                return

            if align is None:
                # first
                align = t.align
            else:
                if t.align != align:
                    return

        return align

    @property
    def size(self):
        single_type = self.get_single_type()

        if single_type is not None:
            return single_type.size

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, value):
        self._tag = value

    @property
    def types(self):
        return self._types

    def __getitem__(self, key):
        return self.types[key]

    def get_single_type(self):
        if len(self.members) == 1:
            return list(self.members.values())[0]

    @property
    def is_dynamic(self):
        single_type = self.get_single_type()

        if single_type is not None:
            return single_type.is_dynamic

        return True


class Trace:
    def __init__(self):
        self._byte_order = None
        self._packet_header_type = None
        self._uuid = None

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def byte_order(self):
        return self._byte_order

    @byte_order.setter
    def byte_order(self, value):
        self._byte_order = value

    @property
    def packet_header_type(self):
        return self._packet_header_type

    @packet_header_type.setter
    def packet_header_type(self, value):
        self._packet_header_type = value


class Env(collections.OrderedDict):
    pass


class Clock:
    def __init__(self):
        self.set_default_name()
        self.set_default_uuid()
        self.set_default_description()
        self.set_default_freq()
        self.set_default_error_cycles()
        self.set_default_offset_seconds()
        self.set_default_offset_cycles()
        self.set_default_absolute()
        self.set_default_return_ctype()

    def set_default_name(self):
        self._name = None

    def set_default_uuid(self):
        self._uuid = None

    def set_default_description(self):
        self._description = None

    def set_default_freq(self):
        self._freq = 1000000000

    def set_default_error_cycles(self):
        self._error_cycles = 0

    def set_default_offset_seconds(self):
        self._offset_seconds = 0

    def set_default_offset_cycles(self):
        self._offset_cycles = 0

    def set_default_absolute(self):
        self._absolute = False

    def set_default_return_ctype(self):
        self._return_ctype = 'uint32_t'

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def error_cycles(self):
        return self._error_cycles

    @error_cycles.setter
    def error_cycles(self, value):
        self._error_cycles = value

    @property
    def freq(self):
        return self._freq

    @freq.setter
    def freq(self, value):
        self._freq = value

    @property
    def offset_seconds(self):
        return self._offset_seconds

    @offset_seconds.setter
    def offset_seconds(self, value):
        self._offset_seconds = value

    @property
    def offset_cycles(self):
        return self._offset_cycles

    @offset_cycles.setter
    def offset_cycles(self, value):
        self._offset_cycles = value

    @property
    def absolute(self):
        return self._absolute

    @absolute.setter
    def absolute(self, value):
        self._absolute = value

    @property
    def return_ctype(self):
        return self._return_ctype

    @return_ctype.setter
    def return_ctype(self, value):
        self._return_ctype = value


LogLevel = collections.namedtuple('LogLevel', ['name', 'value'])


class Event:
    def __init__(self):
        self._id = None
        self._name = None
        self._log_level = None
        self._context_type = None
        self._payload_type = None

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def log_level(self):
        return self._log_level

    @log_level.setter
    def log_level(self, value):
        self._log_level = value

    @property
    def context_type(self):
        return self._context_type

    @context_type.setter
    def context_type(self, value):
        self._context_type = value

    @property
    def payload_type(self):
        return self._payload_type

    @payload_type.setter
    def payload_type(self, value):
        self._payload_type = value

    def __getitem__(self, key):
        if type(self.payload_type) in [Struct, Variant]:
            return self.payload_type[key]

        raise TypeError('{} is not subscriptable')


class Stream:
    def __init__(self):
        self._id = 0
        self._name = None
        self._packet_context_type = None
        self._event_header_type = None
        self._event_context_type = None
        self._events = collections.OrderedDict()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def packet_context_type(self):
        return self._packet_context_type

    @packet_context_type.setter
    def packet_context_type(self, value):
        self._packet_context_type = value

    @property
    def event_header_type(self):
        return self._event_header_type

    @event_header_type.setter
    def event_header_type(self, value):
        self._event_header_type = value

    @property
    def event_context_type(self):
        return self._event_context_type

    @event_context_type.setter
    def event_context_type(self, value):
        self._event_context_type = value

    @property
    def events(self):
        return self._events

    def is_event_empty(self, event):
        total_fields = 0

        if self.event_header_type:
            total_fields += len(self.event_header_type)

        if self.event_context_type:
            total_fields += len(self.event_context_type)

        if event.context_type:
            total_fields += len(event.context_type)

        if event.payload_type:
            total_fields += len(event.payload_type)

        return total_fields == 0


class Metadata:
    def __init__(self):
        self._trace = None
        self._env = collections.OrderedDict()
        self._clocks = collections.OrderedDict()
        self._streams = collections.OrderedDict()
        self._default_stream_name = None

    @property
    def trace(self):
        return self._trace

    @trace.setter
    def trace(self, value):
        self._trace = value

    @property
    def env(self):
        return self._env

    @env.setter
    def env(self, value):
        self._env = value

    @property
    def clocks(self):
        return self._clocks

    @clocks.setter
    def clocks(self, value):
        self._clocks = value

    @property
    def streams(self):
        return self._streams

    @streams.setter
    def streams(self, value):
        self._streams = value

    @property
    def default_stream_name(self):
        return self._default_stream_name

    @default_stream_name.setter
    def default_stream_name(self, value):
        self._default_stream_name = value

    @property
    def default_stream(self):
        if self._default_stream_name in self._streams:
            return self._streams[self._default_stream_name]



# The MIT License (MIT)
#
# Copyright (c) 2015 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class CodeGenerator:
    def __init__(self, indent_string):
        self._indent_string = indent_string
        self.reset()

    @property
    def code(self):
        return '\n'.join(self._lines)

    def reset(self):
        self._lines = []
        self._indent = 0
        self._glue = False

    def add_line(self, line):
        if self._glue:
            self.append_to_last_line(line)
            self._glue = False
            return

        indent_string = self._get_indent_string()
        self._lines.append(indent_string + str(line))

    def add_lines(self, lines):
        if type(lines) is str:
            lines = lines.split('\n')

        for line in lines:
            self.add_line(line)

    def add_glue(self):
        self._glue = True

    def append_to_last_line(self, s):
        if self._lines:
            self._lines[-1] += str(s)

    def add_empty_line(self):
        self._lines.append('')

    def add_cc_line(self, comment):
        self.add_line('/* {} */'.format(comment))

    def append_cc_to_last_line(self, comment, with_space=True):
        if with_space:
            sp = ' '
        else:
            sp = ''

        self.append_to_last_line('{}/* {} */'.format(sp, comment))

    def indent(self):
        self._indent += 1

    def unindent(self):
        self._indent = max(self._indent - 1, 0)

    def _get_indent_string(self):
        return self._indent_string * self._indent



# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pkg_resources import resource_filename
from termcolor import cprint, colored
import barectf.tsdl182gen
import barectf.config
import barectf.gen
import argparse
import os.path
import barectf
import sys
import os
import re


def _perror(msg):
    cprint('Error: ', 'red', end='', file=sys.stderr)
    cprint(msg, 'red', attrs=['bold'], file=sys.stderr)
    sys.exit(1)


def _pconfig_error(e):
    lines = []

    while True:
        if e is None:
            break

        lines.append(str(e))

        if not hasattr(e, 'prev'):
            break

        e = e.prev

    if len(lines) == 1:
        _perror(lines[0])

    cprint('Error:', 'red', file=sys.stderr)

    for i, line in enumerate(lines):
        suf = ':' if i < len(lines) - 1 else ''
        cprint('  ' + line + suf, 'red', attrs=['bold'], file=sys.stderr)

    sys.exit(1)


def _psuccess(msg):
    cprint('{}'.format(msg), 'green', attrs=['bold'])


def _parse_args():
    ap = argparse.ArgumentParser()

    ap.add_argument('-c', '--code-dir', metavar='DIR', action='store',
                    default=os.getcwd(),
                    help='output directory of C source file')
    ap.add_argument('--dump-config', action='store_true',
                    help='also dump the effective YAML configuration file used for generation')
    ap.add_argument('-H', '--headers-dir', metavar='DIR', action='store',
                    default=os.getcwd(),
                    help='output directory of C header files')
    ap.add_argument('-I', '--include-dir', metavar='DIR', action='append',
                    default=[],
                    help='add directory DIR to the list of directories to be searched for include files')
    ap.add_argument('--ignore-include-not-found', action='store_true',
                    help='continue to process the configuration file when included files are not found')
    ap.add_argument('-m', '--metadata-dir', metavar='DIR', action='store',
                    default=os.getcwd(),
                    help='output directory of CTF metadata')
    ap.add_argument('-p', '--prefix', metavar='PREFIX', action='store',
                    help='override configuration\'s prefix')
    ap.add_argument('-V', '--version', action='version',
                    version='%(prog)s {}'.format(barectf.__version__))
    ap.add_argument('config', metavar='CONFIG', action='store',
                    help='barectf YAML configuration file')

    # parse args
    args = ap.parse_args()

    # validate output directories
    for d in [args.code_dir, args.headers_dir, args.metadata_dir] + args.include_dir:
        if not os.path.isdir(d):
            _perror('"{}" is not an existing directory'.format(d))

    # validate that configuration file exists
    if not os.path.isfile(args.config):
        _perror('"{}" is not an existing, regular file'.format(args.config))

    # append current working directory and provided include directory
    args.include_dir += [os.getcwd(), resource_filename(__name__, 'include')]

    return args


def _write_file(d, name, content):
    with open(os.path.join(d, name), 'w') as f:
        f.write(content)


def run():
    # parse arguments
    args = _parse_args()

    # create configuration
    try:
        config = barectf.config.from_yaml_file(args.config, args.include_dir,
                                               args.ignore_include_not_found,
                                               args.dump_config)
    except barectf.config.ConfigError as e:
        _pconfig_error(e)
    except Exception as e:
        _perror('unknown exception: {}'.format(e))

    # replace prefix if needed
    if args.prefix:
        try:
            config.prefix = args.prefix
        except barectf.config.ConfigError as e:
            _pconfig_error(e)

    # generate metadata
    metadata = barectf.tsdl182gen.from_metadata(config.metadata)

    try:
        _write_file(args.metadata_dir, 'metadata', metadata)
    except Exception as e:
        _perror('cannot write metadata file: {}'.format(e))

    # create generator
    generator = barectf.gen.CCodeGenerator(config)

    # generate C headers
    header = generator.generate_header()
    bitfield_header = generator.generate_bitfield_header()

    try:
        _write_file(args.headers_dir, generator.get_header_filename(), header)
        _write_file(args.headers_dir, generator.get_bitfield_header_filename(),
                    bitfield_header)
    except Exception as e:
        _perror('cannot write header files: {}'.format(e))

    # generate C source
    c_src = generator.generate_c_src()

    try:
        _write_file(args.code_dir, '{}.c'.format(config.prefix.rstrip('_')),
                    c_src)
    except Exception as e:
        _perror('cannot write C source file: {}'.format(e))



#!/usr/bin/env python3
#
# The MIT License (MIT)
#
# Copyright (c) 2014-2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from setuptools import setup
import sys


def _check_python3():
  # make sure we run Python 3+ here
  v = sys.version_info

  if v.major < 3:
      sys.stderr.write('Sorry, barectf needs Python 3\n')
      sys.exit(1)


_check_python3()


import barectf


setup(name='barectf',
      version=barectf.__version__,
      description='Generator of ANSI C tracers which output CTF',
      author='Philippe Proulx',
      author_email='eeppeliteloop@gmail.com',
      license='MIT',
      keywords='ctf generator tracing bare-metal bare-machine',
      url='http://barectf.org',
      packages=[
          'barectf',
      ],
      package_data={
          'barectf': [
              'include/*.yaml',
          ],
      },
      install_requires=[
          'termcolor',
          'pyyaml',
      ],
      entry_points={
          'console_scripts': [
              'barectf = barectf.cli:run'
          ],
      })



# The MIT License (MIT)
#
# Copyright (c) 2014-2019 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from barectf import templates
from barectf import metadata
import barectf.codegen
import collections
import itertools
import argparse
import datetime
import barectf
import copy
import sys
import os
import re


def _align(v, align):
    return (v + (align - 1)) & -align


class _SerializationAction:
    def __init__(self, offset_in_byte, type, names):
        assert(offset_in_byte >= 0 and offset_in_byte < 8)
        self._offset_in_byte = offset_in_byte
        self._type = type
        self._names = copy.deepcopy(names)

    @property
    def offset_in_byte(self):
        return self._offset_in_byte

    @property
    def type(self):
        return self._type

    @property
    def names(self):
        return self._names


class _AlignSerializationAction(_SerializationAction):
    def __init__(self, offset_in_byte, type, names, value):
        super().__init__(offset_in_byte, type, names)
        self._value = value

    @property
    def value(self):
        return self._value


class _SerializeSerializationAction(_SerializationAction):
    def __init__(self, offset_in_byte, type, names):
        super().__init__(offset_in_byte, type, names)


class _SerializationActions:
    def __init__(self):
        self.reset()

    def reset(self):
        self._last_alignment = None
        self._last_bit_array_size = None
        self._actions = []
        self._names = []
        self._offset_in_byte = 0

    def append_root_scope_type(self, t, name):
        if t is None:
            return

        assert(type(t) is metadata.Struct)
        self._names = [name]
        self._append_type(t)

    @property
    def actions(self):
        return self._actions

    def align(self, alignment):
        do_align = self._must_align(alignment)
        self._last_alignment = alignment
        self._last_bit_array_size = alignment
        self._try_append_align_action(alignment, do_align)

    def _must_align(self, align_req):
        return self._last_alignment != align_req or self._last_bit_array_size % align_req != 0

    def _append_type(self, t):
        assert(type(t) in (metadata.Struct, metadata.String, metadata.Integer,
                           metadata.FloatingPoint, metadata.Enum,
                           metadata.Array))

        if type(t) in (metadata.String, metadata.Array):
            assert(type(t) is metadata.String or self._names[-1] == 'uuid')
            do_align = self._must_align(8)
            self._last_alignment = 8
            self._last_bit_array_size = 8
            self._try_append_align_action(8, do_align, t)
            self._append_serialize_action(t)
        elif type(t) in (metadata.Integer, metadata.FloatingPoint,
                         metadata.Enum, metadata.Struct):
            do_align = self._must_align(t.align)
            self._last_alignment = t.align

            if type(t) is metadata.Struct:
                self._last_bit_array_size = t.align
            else:
                self._last_bit_array_size = t.size

            self._try_append_align_action(t.align, do_align, t)

            if type(t) is metadata.Struct:
                for field_name, field_type in t.fields.items():
                    self._names.append(field_name)
                    self._append_type(field_type)
                    del self._names[-1]
            else:
                self._append_serialize_action(t, t.size)

    def _try_append_align_action(self, alignment, do_align, t=None):
        offset_in_byte = self._offset_in_byte
        self._offset_in_byte = _align(self._offset_in_byte, alignment) % 8

        if do_align and alignment > 1:
            self._actions.append(_AlignSerializationAction(offset_in_byte,
                                                           t, self._names,
                                                           alignment))

    def _append_serialize_action(self, t, size=None):
        assert(type(t) in (metadata.Integer, metadata.FloatingPoint,
                           metadata.Enum, metadata.String,
                           metadata.Array))

        offset_in_byte = self._offset_in_byte

        if t.size is not None:
            self._offset_in_byte += t.size
            self._offset_in_byte %= 8

        self._actions.append(_SerializeSerializationAction(offset_in_byte,
                                                           t, self._names))


_PREFIX_TPH = 'tph_'
_PREFIX_SPC = 'spc_'
_PREFIX_SEH = 'seh_'
_PREFIX_SEC = 'sec_'
_PREFIX_EC = 'ec_'
_PREFIX_EP = 'ep_'
_PREFIX_TO_NAME = {
    _PREFIX_TPH: 'trace packet header',
    _PREFIX_SPC: 'stream packet context',
    _PREFIX_SEH: 'stream event header',
    _PREFIX_SEC: 'stream event context',
    _PREFIX_EC: 'event context',
    _PREFIX_EP: 'event payload',
}


class CCodeGenerator:
    def __init__(self, cfg):
        self._cfg = cfg
        self._cg = barectf.codegen.CodeGenerator('\t')
        self._type_to_get_ctype_func = {
            metadata.Integer: self._get_int_ctype,
            metadata.FloatingPoint: self._get_float_ctype,
            metadata.Enum: self._get_enum_ctype,
            metadata.String: self._get_string_ctype,
        }
        self._type_to_generate_serialize_func = {
            metadata.Integer: self._generate_serialize_int,
            metadata.FloatingPoint: self._generate_serialize_float,
            metadata.Enum: self._generate_serialize_enum,
            metadata.String: self._generate_serialize_string,
        }
        self._saved_serialization_actions = {}

    def _get_stream_clock(self, stream):
        field = None

        if stream.event_header_type is not None:
            if 'timestamp' in stream.event_header_type.fields:
                field = stream.event_header_type['timestamp']

        if stream.packet_context_type is not None:
            if field is None and 'timestamp_begin' in stream.packet_context_type.fields:
                field = stream.packet_context_type['timestamp_begin']

            if field is None and 'timestamp_end' in stream.packet_context_type.fields:
                field = stream.packet_context_type['timestamp_end']

        if field is None:
            return

        if field.property_mappings:
            return field.property_mappings[0].object

    def _generate_ctx_parent(self):
        tmpl = templates._CTX_PARENT
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix))

    def _generate_ctx(self, stream):
        tmpl = templates._CTX_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name))
        tmpl = 'uint32_t off_tph_{fname};'
        self._cg.indent()
        trace_packet_header_type = self._cfg.metadata.trace.packet_header_type

        if trace_packet_header_type is not None:
            for field_name in trace_packet_header_type.fields:
                self._cg.add_lines(tmpl.format(fname=field_name))

        tmpl = 'uint32_t off_spc_{fname};'

        if stream.packet_context_type is not None:
            for field_name in stream.packet_context_type.fields:
                self._cg.add_lines(tmpl.format(fname=field_name))

        clock = self._get_stream_clock(stream)

        if clock is not None:
            line = '{} cur_last_event_ts;'.format(clock.return_ctype)
            self._cg.add_line(line)

        self._cg.unindent()
        tmpl = templates._CTX_END
        self._cg.add_lines(tmpl)

    def _generate_ctxs(self):
        for stream in self._cfg.metadata.streams.values():
            self._generate_ctx(stream)

    def _generate_clock_cb(self, clock):
        tmpl = templates._CLOCK_CB
        self._cg.add_lines(tmpl.format(return_ctype=clock.return_ctype,
                                       cname=clock.name))

    def _generate_clock_cbs(self):
        for clock in self._cfg.metadata.clocks.values():
            self._generate_clock_cb(clock)

    def _generate_platform_callbacks(self):
        tmpl = templates._PLATFORM_CALLBACKS_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix))
        self._cg.indent()
        self._generate_clock_cbs()
        self._cg.unindent()
        tmpl = templates._PLATFORM_CALLBACKS_END
        self._cg.add_lines(tmpl)

    def generate_bitfield_header(self):
        self._cg.reset()
        tmpl = templates._BITFIELD
        tmpl = tmpl.replace('$prefix$', self._cfg.prefix)
        tmpl = tmpl.replace('$PREFIX$', self._cfg.prefix.upper())

        if self._cfg.metadata.trace.byte_order == metadata.ByteOrder.BE:
            endian_def = 'BIG_ENDIAN'
        else:
            endian_def = 'LITTLE_ENDIAN'

        tmpl = tmpl.replace('$ENDIAN_DEF$', endian_def)
        self._cg.add_lines(tmpl)

        return self._cg.code

    def _generate_func_init_proto(self):
        tmpl = templates._FUNC_INIT_PROTO
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix))

    def _get_int_ctype(self, t):
        signed = 'u' if not t.signed else ''

        if t.size <= 8:
            sz = '8'
        elif t.size <= 16:
            sz = '16'
        elif t.size <= 32:
            sz = '32'
        elif t.size == 64:
            sz = '64'

        return '{}int{}_t'.format(signed, sz)

    def _get_float_ctype(self, t):
        if t.exp_size == 8 and t.mant_size == 24 and t.align == 32:
            ctype = 'float'
        elif t.exp_size == 11 and t.mant_size == 53 and t.align == 64:
            ctype = 'double'
        else:
            ctype = 'uint64_t'

        return ctype

    def _get_enum_ctype(self, t):
        return self._get_int_ctype(t.value_type)

    def _get_string_ctype(self, t):
        return 'const char *'

    def _get_type_ctype(self, t):
        return self._type_to_get_ctype_func[type(t)](t)

    def _generate_type_ctype(self, t):
        ctype = self._get_type_ctype(t)
        self._cg.append_to_last_line(ctype)

    def _generate_proto_param(self, t, name):
        self._generate_type_ctype(t)
        self._cg.append_to_last_line(' ')
        self._cg.append_to_last_line(name)

    def _generate_proto_params(self, t, name_prefix, exclude_list):
        self._cg.indent()

        for field_name, field_type in t.fields.items():
            if field_name in exclude_list:
                continue

            name = name_prefix + field_name
            self._cg.append_to_last_line(',')
            self._cg.add_line('')
            self._generate_proto_param(field_type, name)

        self._cg.unindent()

    def _generate_func_open_proto(self, stream):
        tmpl = templates._FUNC_OPEN_PROTO_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name))
        trace_packet_header_type = self._cfg.metadata.trace.packet_header_type

        if trace_packet_header_type is not None:
            exclude_list = ['magic', 'stream_id', 'uuid']
            self._generate_proto_params(trace_packet_header_type, _PREFIX_TPH,
                                        exclude_list)

        if stream.packet_context_type is not None:
            exclude_list = [
                'timestamp_begin',
                'timestamp_end',
                'packet_size',
                'content_size',
                'events_discarded',
            ]
            self._generate_proto_params(stream.packet_context_type,
                                        _PREFIX_SPC, exclude_list)

        tmpl = templates._FUNC_OPEN_PROTO_END
        self._cg.add_lines(tmpl)

    def _generate_func_close_proto(self, stream):
        tmpl = templates._FUNC_CLOSE_PROTO
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name))

    def _generate_func_trace_proto_params(self, stream, event):
        if stream.event_header_type is not None:
            exclude_list = [
                'id',
                'timestamp',
            ]
            self._generate_proto_params(stream.event_header_type,
                                        _PREFIX_SEH, exclude_list)

        if stream.event_context_type is not None:
            self._generate_proto_params(stream.event_context_type,
                                        _PREFIX_SEC, [])

        if event.context_type is not None:
            self._generate_proto_params(event.context_type,
                                        _PREFIX_EC, [])

        if event.payload_type is not None:
            self._generate_proto_params(event.payload_type,
                                        _PREFIX_EP, [])

    def _generate_func_trace_proto(self, stream, event):
        tmpl = templates._FUNC_TRACE_PROTO_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name, evname=event.name))
        self._generate_func_trace_proto_params(stream, event)
        tmpl = templates._FUNC_TRACE_PROTO_END
        self._cg.add_lines(tmpl)

    def _punctuate_proto(self):
        self._cg.append_to_last_line(';')

    def generate_header(self):
        self._cg.reset()
        dt = datetime.datetime.now().isoformat()
        bh_filename = self.get_bitfield_header_filename()
        prefix_def = ''
        default_stream_def = ''

        if self._cfg.options.gen_prefix_def:
            prefix_def = '#define _BARECTF_PREFIX {}'.format(self._cfg.prefix)

        if self._cfg.options.gen_default_stream_def and self._cfg.metadata.default_stream_name is not None:
            default_stream_def = '#define _BARECTF_DEFAULT_STREAM {}'.format(self._cfg.metadata.default_stream_name)

        default_stream_trace_defs = ''
        default_stream_name = self._cfg.metadata.default_stream_name

        if default_stream_name is not None:
            default_stream = self._cfg.metadata.streams[default_stream_name]
            lines = []

            for ev_name in default_stream.events.keys():
                tmpl = templates._DEFINE_DEFAULT_STREAM_TRACE
                define = tmpl.format(prefix=self._cfg.prefix,
                                     sname=default_stream_name,
                                     evname=ev_name)
                lines.append(define)

            default_stream_trace_defs = '\n'.join(lines)

        tmpl = templates._HEADER_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       ucprefix=self._cfg.prefix.upper(),
                                       bitfield_header_filename=bh_filename,
                                       version=barectf.__version__, date=dt,
                                       prefix_def=prefix_def,
                                       default_stream_def=default_stream_def,
                                       default_stream_trace_defs=default_stream_trace_defs))
        self._cg.add_empty_line()

        # platform callbacks structure
        self._generate_platform_callbacks()
        self._cg.add_empty_line()

        # context parent
        self._generate_ctx_parent()
        self._cg.add_empty_line()

        # stream contexts
        self._generate_ctxs()
        self._cg.add_empty_line()

        # initialization function prototype
        self._generate_func_init_proto()
        self._punctuate_proto()
        self._cg.add_empty_line()

        for stream in self._cfg.metadata.streams.values():
            self._generate_func_open_proto(stream)
            self._punctuate_proto()
            self._cg.add_empty_line()
            self._generate_func_close_proto(stream)
            self._punctuate_proto()
            self._cg.add_empty_line()

            for ev in stream.events.values():
                self._generate_func_trace_proto(stream, ev)
                self._punctuate_proto()
                self._cg.add_empty_line()

        tmpl = templates._HEADER_END
        self._cg.add_lines(tmpl.format(ucprefix=self._cfg.prefix.upper()))

        return self._cg.code

    def _get_call_event_param_list_from_struct(self, t, prefix, exclude_list):
        lst = ''

        for field_name in t.fields:
            if field_name in exclude_list:
                continue

            lst += ', {}{}'.format(prefix, field_name)

        return lst

    def _get_call_event_param_list(self, stream, event):
        lst = ''
        gcp_func = self._get_call_event_param_list_from_struct

        if stream.event_header_type is not None:
            exclude_list = [
                'id',
                'timestamp',
            ]
            lst += gcp_func(stream.event_header_type, _PREFIX_SEH, exclude_list)

        if stream.event_context_type is not None:
            lst += gcp_func(stream.event_context_type, _PREFIX_SEC, [])

        if event.context_type is not None:
            lst += gcp_func(event.context_type, _PREFIX_EC, [])

        if event.payload_type is not None:
            lst += gcp_func(event.payload_type, _PREFIX_EP, [])

        return lst

    def _generate_align(self, at, align):
        self._cg.add_line('_ALIGN({}, {});'.format(at, align))

    def _generate_align_type(self, at, t):
        if t.align == 1:
            return

        self._generate_align(at, t.align)

    def _generate_incr_pos(self, var, value):
        self._cg.add_line('{} += {};'.format(var, value))

    def _generate_incr_pos_bytes(self, var, value):
        self._generate_incr_pos(var, '_BYTES_TO_BITS({})'.format(value))

    def _generate_func_get_event_size_proto(self, stream, event):
        tmpl = templates._FUNC_GET_EVENT_SIZE_PROTO_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name, evname=event.name))
        self._generate_func_trace_proto_params(stream, event)
        tmpl = templates._FUNC_GET_EVENT_SIZE_PROTO_END
        self._cg.add_lines(tmpl)

    def _generate_func_get_event_size(self, stream, event):
        self._generate_func_get_event_size_proto(stream, event)
        tmpl = templates._FUNC_GET_EVENT_SIZE_BODY_BEGIN
        lines = tmpl.format(prefix=self._cfg.prefix)
        self._cg.add_lines(lines)
        self._cg.add_empty_line()
        self._cg.indent()
        ser_actions = _SerializationActions()
        ser_actions.append_root_scope_type(stream.event_header_type,
                                            _PREFIX_SEH)
        ser_actions.append_root_scope_type(stream.event_context_type,
                                            _PREFIX_SEC)
        ser_actions.append_root_scope_type(event.context_type, _PREFIX_EC)
        ser_actions.append_root_scope_type(event.payload_type, _PREFIX_EP)

        for action in ser_actions.actions:
            if type(action) is _AlignSerializationAction:
                if action.names:
                    if len(action.names) == 1:
                        line = 'align {} structure'.format(_PREFIX_TO_NAME[action.names[0]])
                    else:
                        fmt = 'align field "{}" ({})'
                        line = fmt.format(action.names[-1],
                                          _PREFIX_TO_NAME[action.names[0]])

                    self._cg.add_cc_line(line)

                self._generate_align('at', action.value)
                self._cg.add_empty_line()
            elif type(action) is _SerializeSerializationAction:
                assert(len(action.names) >= 2)
                fmt = 'add size of field "{}" ({})'
                line = fmt.format(action.names[-1], _PREFIX_TO_NAME[action.names[0]])
                self._cg.add_cc_line(line)

                if type(action.type) is metadata.String:
                    param = ''.join(action.names)
                    self._generate_incr_pos_bytes('at',
                                                  'strlen({}) + 1'.format(param))
                else:
                    self._generate_incr_pos('at', action.type.size)

                self._cg.add_empty_line()

        self._cg.unindent()
        tmpl = templates._FUNC_GET_EVENT_SIZE_BODY_END
        self._cg.add_lines(tmpl)

    def _generate_func_serialize_event_proto(self, stream, event):
        tmpl = templates._FUNC_SERIALIZE_EVENT_PROTO_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name, evname=event.name))
        self._generate_func_trace_proto_params(stream, event)
        tmpl = templates._FUNC_SERIALIZE_EVENT_PROTO_END
        self._cg.add_lines(tmpl)

    def _generate_bitfield_write(self, ctype, var, ctx, action):
        ptr = '&{ctx}->buf[_BITS_TO_BYTES({ctx}->at)]'.format(ctx=ctx)
        start = action.offset_in_byte
        suffix = 'le' if action.type.byte_order is metadata.ByteOrder.LE else 'be'
        func = '{}bt_bitfield_write_{}'.format(self._cfg.prefix, suffix)
        call_fmt = '{func}({ptr}, uint8_t, {start}, {size}, {ctype}, ({ctype}) {var});'
        call = call_fmt.format(func=func, ptr=ptr, start=start,
                               size=action.type.size, ctype=ctype, var=var)
        self._cg.add_line(call)

    def _generate_serialize_int(self, var, ctx, action):
        ctype = self._get_int_ctype(action.type)
        self._generate_bitfield_write(ctype, var, ctx, action)
        self._generate_incr_pos('{}->at'.format(ctx), action.type.size)

    def _generate_serialize_float(self, var, ctx, action):
        ctype = self._get_type_ctype(action.type)
        flt_dbl = False

        if ctype == 'float' or ctype == 'double':
            flt_dbl = True

            if ctype == 'float':
                union_name = 'f2u'
                int_ctype = 'uint32_t'
            elif ctype == 'double':
                union_name = 'd2u'
                int_ctype = 'uint64_t'

            # union for reading the bytes of the floating point number
            self._cg.add_empty_line()
            self._cg.add_line('{')
            self._cg.indent()
            self._cg.add_line('union {name} {name};'.format(name=union_name))
            self._cg.add_empty_line()
            self._cg.add_line('{}.f = {};'.format(union_name, var))
            bf_var = '{}.u'.format(union_name)
        else:
            bf_var = '({}) {}'.format(ctype, var)
            int_ctype = ctype

        self._generate_bitfield_write(int_ctype, bf_var, ctx, action)

        if flt_dbl:
            self._cg.unindent()
            self._cg.add_line('}')
            self._cg.add_empty_line()

        self._generate_incr_pos('{}->at'.format(ctx), action.type.size)

    def _generate_serialize_enum(self, var, ctx, action):
        sub_action = _SerializeSerializationAction(action.offset_in_byte,
                                                   action.type.value_type,
                                                   action.names)
        self._generate_serialize_from_action(var, ctx, sub_action)

    def _generate_serialize_string(self, var, ctx, action):
        tmpl = '_write_cstring({}, {});'.format(ctx, var)
        self._cg.add_lines(tmpl)

    def _generate_serialize_from_action(self, var, ctx, action):
        func = self._type_to_generate_serialize_func[type(action.type)]
        func(var, ctx, action)

    def _generate_serialize_statements_from_actions(self, prefix, action_iter,
                                                    spec_src=None):
        for action in action_iter:
            if type(action) is _AlignSerializationAction:
                if action.names:
                    if len(action.names) == 1:
                        line = 'align {} structure'.format(_PREFIX_TO_NAME[action.names[0]])
                    else:
                        fmt = 'align field "{}" ({})'
                        line = fmt.format(action.names[-1],
                                          _PREFIX_TO_NAME[action.names[0]])

                    self._cg.add_cc_line(line)

                self._generate_align('ctx->at', action.value)
                self._cg.add_empty_line()
            elif type(action) is _SerializeSerializationAction:
                assert(len(action.names) >= 2)
                fmt = 'serialize field "{}" ({})'
                line = fmt.format(action.names[-1],
                                  _PREFIX_TO_NAME[action.names[0]])
                self._cg.add_cc_line(line)
                field_name = action.names[-1]
                src = prefix + field_name

                if spec_src is not None and field_name in spec_src:
                    src = spec_src[field_name]

                self._generate_serialize_from_action(src, 'ctx', action)
                self._cg.add_empty_line()

    def _generate_func_serialize_event(self, stream, event, orig_ser_actions):
        self._generate_func_serialize_event_proto(stream, event)
        tmpl = templates._FUNC_SERIALIZE_EVENT_BODY_BEGIN
        lines = tmpl.format(prefix=self._cfg.prefix)
        self._cg.add_lines(lines)
        self._cg.indent()
        self._cg.add_empty_line()

        if stream.event_header_type is not None:
            t = stream.event_header_type
            exclude_list = ['timestamp', 'id']
            params = self._get_call_event_param_list_from_struct(t, _PREFIX_SEH,
                                                                 exclude_list)
            tmpl = '_serialize_stream_event_header_{sname}(ctx, {evid}{params});'
            self._cg.add_cc_line('stream event header')
            self._cg.add_line(tmpl.format(sname=stream.name, evid=event.id,
                                          params=params))
            self._cg.add_empty_line()

        if stream.event_context_type is not None:
            t = stream.event_context_type
            params = self._get_call_event_param_list_from_struct(t, _PREFIX_SEC,
                                                                 [])
            tmpl = '_serialize_stream_event_context_{sname}(ctx{params});'
            self._cg.add_cc_line('stream event context')
            self._cg.add_line(tmpl.format(sname=stream.name, params=params))
            self._cg.add_empty_line()

        if event.context_type is not None or event.payload_type is not None:
            ser_actions = copy.deepcopy(orig_ser_actions)

        if event.context_type is not None:
            ser_action_index = len(ser_actions.actions)
            ser_actions.append_root_scope_type(event.context_type, _PREFIX_EC)
            ser_action_iter = itertools.islice(ser_actions.actions,
                                               ser_action_index, None)
            self._generate_serialize_statements_from_actions(_PREFIX_EC,
                                                             ser_action_iter)

        if event.payload_type is not None:
            ser_action_index = len(ser_actions.actions)
            ser_actions.append_root_scope_type(event.payload_type, _PREFIX_EP)
            ser_action_iter = itertools.islice(ser_actions.actions,
                                               ser_action_index, None)
            self._generate_serialize_statements_from_actions(_PREFIX_EP,
                                                             ser_action_iter)

        self._cg.unindent()
        tmpl = templates._FUNC_SERIALIZE_EVENT_BODY_END
        self._cg.add_lines(tmpl)

    def _generate_func_serialize_stream_event_header_proto(self, stream):
        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_HEADER_PROTO_BEGIN
        clock_ctype = 'const int'
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name))

        if stream.event_header_type is not None:
            exclude_list = [
                'id',
                'timestamp',
            ]
            self._generate_proto_params(stream.event_header_type,
                                        _PREFIX_SEH, exclude_list)

        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_HEADER_PROTO_END
        self._cg.add_lines(tmpl)

    def _generate_func_serialize_stream_event_context_proto(self, stream):
        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_PROTO_BEGIN
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       sname=stream.name))

        if stream.event_context_type is not None:
            self._generate_proto_params(stream.event_context_type,
                                        _PREFIX_SEC, [])

        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_PROTO_END
        self._cg.add_lines(tmpl)

    def _generate_func_serialize_stream_event_header(self, stream,
                                                     ser_action_iter):
        self._generate_func_serialize_stream_event_header_proto(stream)
        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_HEADER_BODY_BEGIN
        lines = tmpl.format(prefix=self._cfg.prefix, sname=stream.name)
        self._cg.add_lines(lines)
        self._cg.indent()
        clock = self._get_stream_clock(stream)

        if clock is not None:
            tmpl = 'struct {prefix}{sname}_ctx *s_ctx = FROM_VOID_PTR(struct {prefix}{sname}_ctx, vctx);'
            line = tmpl.format(prefix=self._cfg.prefix,
                               sname=stream.name)
            self._cg.add_line(line)
            tmpl = 'const {} ts = s_ctx->cur_last_event_ts;'
            line = tmpl.format(clock.return_ctype)
            self._cg.add_line(line)

        self._cg.add_empty_line()

        if stream.event_header_type is not None:
            spec_src = {}

            if 'id' in stream.event_header_type.fields:
                id_t = stream.event_header_type.fields['id']
                id_t_ctype = self._get_int_ctype(id_t)
                spec_src['id'] = '({}) event_id'.format(id_t_ctype)

            if 'timestamp' in stream.event_header_type.fields:
                field = stream.event_header_type.fields['timestamp']
                ts_ctype = self._get_int_ctype(field)
                spec_src['timestamp'] = '({}) ts'.format(ts_ctype)

            self._generate_serialize_statements_from_actions(_PREFIX_SEH,
                                                             ser_action_iter,
                                                             spec_src)

        self._cg.unindent()
        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_HEADER_BODY_END
        self._cg.add_lines(tmpl)

    def _generate_func_serialize_stream_event_context(self, stream,
                                                      ser_action_iter):
        self._generate_func_serialize_stream_event_context_proto(stream)
        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_BODY_BEGIN
        lines = tmpl.format(prefix=self._cfg.prefix)
        self._cg.add_lines(lines)
        self._cg.indent()

        if stream.event_context_type is not None:
            self._generate_serialize_statements_from_actions(_PREFIX_SEC,
                                                             ser_action_iter)

        self._cg.unindent()
        tmpl = templates._FUNC_SERIALIZE_STREAM_EVENT_CONTEXT_BODY_END
        self._cg.add_lines(tmpl)

    def _generate_func_trace(self, stream, event):
        self._generate_func_trace_proto(stream, event)
        params = self._get_call_event_param_list(stream, event)
        clock = self._get_stream_clock(stream)

        if clock is not None:
            tmpl = 'ctx->cur_last_event_ts = ctx->parent.cbs.{}_clock_get_value(ctx->parent.data);'
            save_ts_line = tmpl.format(clock.name)
        else:
            save_ts_line = '/* (no clock) */'

        tmpl = templates._FUNC_TRACE_BODY
        self._cg.add_lines(tmpl.format(sname=stream.name, evname=event.name,
                                       params=params, save_ts=save_ts_line))

    def _generate_func_init(self):
        self._generate_func_init_proto()
        tmpl = templates._FUNC_INIT_BODY
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix))

    def _generate_field_name_cc_line(self, field_name):
        self._cg.add_cc_line('"{}" field'.format(field_name))

    def _save_serialization_action(self, name, action):
        self._saved_serialization_actions[name] = action

    def _get_open_close_ts_line(self, stream):
        clock = self._get_stream_clock(stream)

        if clock is None:
            return ''

        tmpl = '\tconst {} ts = ctx->parent.use_cur_last_event_ts ? ctx->cur_last_event_ts : ctx->parent.cbs.{}_clock_get_value(ctx->parent.data);'
        line = tmpl.format(clock.return_ctype, clock.name)
        return line

    def _generate_func_open(self, stream):
        def generate_save_offset(name, action):
            tmpl = 'ctx->off_spc_{} = ctx->parent.at;'.format(name)
            self._cg.add_line(tmpl)
            self._save_serialization_action(name, action)

        self._generate_func_open_proto(stream)
        tmpl = templates._FUNC_OPEN_BODY_BEGIN
        spc_type = stream.packet_context_type
        ts_line = self._get_open_close_ts_line(stream)
        lines = tmpl.format(ts=ts_line)
        self._cg.add_lines(lines)
        self._cg.indent()
        self._cg.add_cc_line('do not open a packet that is already open')
        self._cg.add_line('if (ctx->parent.packet_is_open) {')
        self._cg.indent()
        self._cg.add_line('ctx->parent.in_tracing_section = saved_in_tracing_section;')
        self._cg.add_line('return;')
        self._cg.unindent()
        self._cg.add_line('}')
        self._cg.add_empty_line()
        self._cg.add_line('ctx->parent.at = 0;')
        tph_type = self._cfg.metadata.trace.packet_header_type
        ser_actions = _SerializationActions()

        if tph_type is not None:
            self._cg.add_empty_line()
            self._cg.add_cc_line('trace packet header')
            self._cg.add_line('{')
            self._cg.indent()
            ser_actions.append_root_scope_type(tph_type, _PREFIX_TPH)

            for action in ser_actions.actions:
                if type(action) is _AlignSerializationAction:
                    if action.names:
                        if len(action.names) == 1:
                            line = 'align trace packet header structure'
                        else:
                            line = 'align field "{}"'.format(action.names[-1])

                        self._cg.add_cc_line(line)

                    self._generate_align('ctx->parent.at', action.value)
                    self._cg.add_empty_line()
                elif type(action) is _SerializeSerializationAction:
                    assert(len(action.names) >= 2)
                    fmt = 'serialize field "{}"'
                    line = fmt.format(action.names[-1])
                    self._cg.add_cc_line(line)
                    field_name = action.names[-1]
                    src = _PREFIX_TPH + field_name

                    if field_name == 'magic':
                        src = '0xc1fc1fc1UL'
                    elif field_name == 'stream_id':
                        stream_id_ctype = self._get_int_ctype(action.type)
                        src = '({}) {}'.format(stream_id_ctype, stream.id)
                    elif field_name == 'uuid':
                        self._cg.add_line('{')
                        self._cg.indent()
                        self._cg.add_line('static uint8_t uuid[] = {')
                        self._cg.indent()

                        for b in self._cfg.metadata.trace.uuid.bytes:
                            self._cg.add_line('{},'.format(b))

                        self._cg.unindent()
                        self._cg.add_line('};')
                        self._cg.add_empty_line()
                        self._generate_align('ctx->parent.at', 8)
                        line = 'memcpy(&ctx->parent.buf[_BITS_TO_BYTES(ctx->parent.at)], uuid, 16);'
                        self._cg.add_line(line)
                        self._generate_incr_pos_bytes('ctx->parent.at', 16)
                        self._cg.unindent()
                        self._cg.add_line('}')
                        self._cg.add_empty_line()
                        continue

                    self._generate_serialize_from_action(src, '(&ctx->parent)', action)
                    self._cg.add_empty_line()

            self._cg.unindent()
            self._cg.add_lines('}')

        spc_action_index = len(ser_actions.actions)

        if spc_type is not None:
            self._cg.add_empty_line()
            self._cg.add_cc_line('stream packet context')
            self._cg.add_line('{')
            self._cg.indent()
            ser_actions.append_root_scope_type(spc_type, _PREFIX_SPC)
            tmpl_off = 'off_spc_{fname}'

            for action in itertools.islice(ser_actions.actions, spc_action_index, None):
                if type(action) is _AlignSerializationAction:
                    if action.names:
                        if len(action.names) == 1:
                            line = 'align stream packet context structure'
                        else:
                            line = 'align field "{}"'.format(action.names[-1])

                        self._cg.add_cc_line(line)

                    self._generate_align('ctx->parent.at', action.value)
                    self._cg.add_empty_line()
                elif type(action) is _SerializeSerializationAction:
                    assert(len(action.names) >= 2)
                    fmt = 'serialize field "{}"'
                    line = fmt.format(action.names[-1])
                    self._cg.add_cc_line(line)
                    field_name = action.names[-1]
                    src = _PREFIX_SPC + field_name
                    skip_int = False

                    if field_name == 'timestamp_begin':
                        ctype = self._get_type_ctype(action.type)
                        src = '({}) ts'.format(ctype)
                    elif field_name in ['timestamp_end', 'content_size',
                                        'events_discarded']:
                        skip_int = True
                    elif field_name == 'packet_size':
                        ctype = self._get_type_ctype(action.type)
                        src = '({}) ctx->parent.packet_size'.format(ctype)

                    if skip_int:
                        generate_save_offset(field_name, action)
                        self._generate_incr_pos('ctx->parent.at',
                                                action.type.size)
                    else:
                        self._generate_serialize_from_action(src, '(&ctx->parent)',
                                                      action)

                    self._cg.add_empty_line()

            self._cg.unindent()
            self._cg.add_lines('}')

        self._cg.unindent()
        tmpl = templates._FUNC_OPEN_BODY_END
        self._cg.add_lines(tmpl)

    def _generate_func_close(self, stream):
        def generate_goto_offset(name):
            tmpl = 'ctx->parent.at = ctx->off_spc_{};'.format(name)
            self._cg.add_line(tmpl)

        self._generate_func_close_proto(stream)
        tmpl = templates._FUNC_CLOSE_BODY_BEGIN
        spc_type = stream.packet_context_type
        ts_line = self._get_open_close_ts_line(stream)
        lines = tmpl.format(ts=ts_line)
        self._cg.add_lines(lines)
        self._cg.indent()
        self._cg.add_cc_line('do not close a packet that is not open')
        self._cg.add_line('if (!ctx->parent.packet_is_open) {')
        self._cg.indent()
        self._cg.add_line('ctx->parent.in_tracing_section = saved_in_tracing_section;')
        self._cg.add_line('return;')
        self._cg.unindent()
        self._cg.add_line('}')
        self._cg.add_empty_line()
        self._cg.add_cc_line('save content size')
        self._cg.add_line('ctx->parent.content_size = ctx->parent.at;')

        if spc_type is not None:
            field_name = 'timestamp_end'

            if field_name in spc_type.fields:
                t = spc_type.fields[field_name]
                ctype = self._get_type_ctype(t)
                src = '({}) ts'.format(ctype)
                self._cg.add_empty_line()
                self._generate_field_name_cc_line(field_name)
                generate_goto_offset(field_name)
                action = self._saved_serialization_actions[field_name]
                self._generate_serialize_from_action(src, '(&ctx->parent)', action)

            field_name = 'content_size'

            if 'content_size' in spc_type.fields:
                t = spc_type.fields[field_name]
                ctype = self._get_type_ctype(t)
                src = '({}) ctx->parent.content_size'.format(ctype)
                self._cg.add_empty_line()
                self._generate_field_name_cc_line(field_name)
                generate_goto_offset(field_name)
                action = self._saved_serialization_actions[field_name]
                self._generate_serialize_from_action(src, '(&ctx->parent)', action)

            field_name = 'events_discarded'

            if field_name in spc_type.fields:
                t = spc_type.fields[field_name]
                ctype = self._get_type_ctype(t)
                src = '({}) ctx->parent.events_discarded'.format(ctype)
                self._cg.add_empty_line()
                self._generate_field_name_cc_line(field_name)
                generate_goto_offset(field_name)
                action = self._saved_serialization_actions[field_name]
                self._generate_serialize_from_action(src, '(&ctx->parent)', action)

        self._cg.unindent()
        tmpl = templates._FUNC_CLOSE_BODY_END
        self._cg.add_lines(tmpl)

    def generate_c_src(self):
        self._cg.reset()
        dt = datetime.datetime.now().isoformat()
        header_filename = self.get_header_filename()
        tmpl = templates._C_SRC
        self._cg.add_lines(tmpl.format(prefix=self._cfg.prefix,
                                       header_filename=header_filename,
                                       version=barectf.__version__, date=dt))
        self._cg.add_empty_line()

        # initialization function
        self._generate_func_init()
        self._cg.add_empty_line()

        for stream in self._cfg.metadata.streams.values():
            self._generate_func_open(stream)
            self._cg.add_empty_line()
            self._generate_func_close(stream)
            self._cg.add_empty_line()
            ser_actions = _SerializationActions()

            if stream.event_header_type is not None:
                ser_actions.append_root_scope_type(stream.event_header_type,
                                                   _PREFIX_SEH)
                self._generate_func_serialize_stream_event_header(stream,
                                                                  iter(ser_actions.actions))
                self._cg.add_empty_line()

            if stream.event_context_type is not None:
                ser_action_index = len(ser_actions.actions)
                ser_actions.append_root_scope_type(stream.event_context_type,
                                                   _PREFIX_SEC)
                ser_action_iter = itertools.islice(ser_actions.actions,
                                                   ser_action_index, None)
                self._generate_func_serialize_stream_event_context(stream,
                                                                   ser_action_iter)
                self._cg.add_empty_line()

            for ev in stream.events.values():
                self._generate_func_get_event_size(stream, ev)
                self._cg.add_empty_line()
                self._generate_func_serialize_event(stream, ev, ser_actions)
                self._cg.add_empty_line()
                self._generate_func_trace(stream, ev)
                self._cg.add_empty_line()

        return self._cg.code

    def get_header_filename(self):
        return '{}.h'.format(self._cfg.prefix.rstrip('_'))

    def get_bitfield_header_filename(self):
        return '{}-bitfield.h'.format(self._cfg.prefix.rstrip('_'))
